from __future__ import annotations

import http
import json
import logging
import sys
from datetime import datetime, timedelta
from urllib.parse import quote

import requests
from bs4 import BeautifulSoup
from furl import furl

from imow.common.actions import IMowActions
from imow.common.consts import IMOW_OAUTH_URI, IMOW_API_URI
from imow.common.exceptions import LoginError
from imow.common.mowerstate import MowerState
from imow.common.mowertask import MowerTask
from imow.common.package_descriptions import *

logger = logging.getLogger('imow')

try:
    assert sys.version_info >= (int(python_major), int(python_minor))
except AssertionError:
    raise RuntimeError(
        f"{package_name!r} requires Python {python_major}.{python_minor}+ (You have Python {sys.version})")


class IMowApi:

    def __init__(self, email: str = None, password: str = None, token: str = None) -> None:

        if (email is None or password is None) and token is None:
            raise LookupError("No email, password or token provided")

        self.http_session = None
        self.csrf_token = None
        self.requestId = None
        self.access_token = None
        self.token_expires = None
        if token:
            logger.debug("New instance: Use Token")
            self.access_token = token
        if not self.access_token:
            logger.debug("New instance: Authenticate with email and password")
            self.get_token(email, password)

    def get_token(self, email: str = "", password: str = "", force_reauth=False) -> (str, datetime):
        """
        look for a token, if present, return. Else authenticate and store new token
        :param email: stihl webapp login email non-url-encoded
        :param password: stihl webapp login password
        :param force_reauth: Force a re-authentication with username and password
        :return: tuple, the access token and a datetime object containing the expire date
        """

        if not self.access_token or force_reauth:
            if force_reauth:
                self.http_session = None
                self.csrf_token = None
                self.requestId = None
                self.access_token = None
                self.token_expires = None

            self.__authenticate(email, password)
            logger.debug("Get Token: Re-Authenticate")

        return self.access_token, self.token_expires

    def __authenticate(self, email: str, password: str) -> [str, str, requests.Response]:
        """
        try the authentication request with fetched csrf and requestId payload
        :param email: stihl webapp login email non-url-encoded
        :param password: stihl webapp login password
        :return: the newly created access token, and expire time besides the legacy response
        """
        self.__fetch_new_csrf_token_and_request_id()
        url = f"{IMOW_OAUTH_URI}/authentication/authenticate/?lang=de"
        encoded_mail = quote(email)
        payload = f'mail={encoded_mail}&password={password}&csrf-token={self.csrf_token}&requestId={self.requestId}'
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
        }
        response = self.api_request(url, "POST", payload=payload, headers=headers)
        if not response.status_code == http.HTTPStatus.OK:
            logger.error(f"Authenticate: {response.status_code} {response.reason}")
            raise ConnectionError(f"{response.status_code} {response.reason}")

        response_url_query_args = furl(response.url).fragment.args
        if "access_token" not in response_url_query_args:
            raise LoginError("STIHL iMow did not return an access_token, check your credentials")

        self.access_token = response_url_query_args["access_token"]
        self.token_expires = datetime.now() + timedelta(seconds=int(response_url_query_args["expires_in"]))
        return self.access_token, self.token_expires, response

    def __fetch_new_csrf_token_and_request_id(self) -> [str, str]:
        """
        Fetch a new csrf_token and requestId to do the authentication as expected by the api
        csrf_token and requestId are used as payload within authentication
        """

        # URL needs whole redirect query parameter
        url = f"{IMOW_OAUTH_URI}/authentication/?lang=de_DE&authorizationRedirectUrl=https%3A%2F%2Foauth2" \
              ".imow.stihl.com%2Fauthorization%2F%3Fresponse_type%3Dtoken%26client_id%3D9526273B-1477-47C6-801C" \
              "-4356F58EF883%26redirect_uri%3Dhttps%253A%252F%252Fapp.imow.stihl.com%252F%2523%252Fauthorize%26state"

        response = self.api_request(url, "GET")

        soup = BeautifulSoup(response.text, 'html.parser')
        try:
            upstream_csrf_token = soup.find("input", {'name': 'csrf-token'}).get('value')
            upstream_request_id = soup.find("input", {'name': 'requestId'}).get('value')
        except AttributeError:
            raise ProcessLookupError("Did not found necessary csrf token and/or request id in html source")

        self.csrf_token = upstream_csrf_token
        self.requestId = upstream_request_id
        logger.debug("CSRF: new token and request id <Redacted>")
        return self.csrf_token, self.requestId

    def api_request(self, url, method, payload=None, headers=None) -> requests.Response:
        """
        Do a standardized request against the stihl imow webapi, with predefined headers
        :param url: The target URL
        :param method: The Method to use
        :param payload: optional payload
        :param headers: optional update headers
        :return: the requests.Response
        """
        if not self.http_session:
            self.http_session = requests.Session()

        if payload is None:
            payload = {}

        headers_obj = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:88.0) Gecko/20100101 Firefox/88.0',
            'Accept': 'application/json, text/plain, */*',
            'Accept-Language': 'de,en-US;q=0.7,en;q=0.3',
            'Authorization': f'Bearer {self.access_token if self.access_token else ""}',
            'Origin': 'https://app.imow.stihl.com',
            'DNT': '1',
            'Connection': 'keep-alive',
            'Referer': 'https://app.imow.stihl.com/',
            'Pragma': 'no-cache',
            'Cache-Control': 'no-cache',
            'TE': 'Trailers',
            'Content-Type': 'application/json'
        }
        if headers:
            headers_obj.update(headers)

        response = self.http_session.request(method, url, headers=headers_obj, data=payload)
        self.http_session.close()
        if response.status_code not in (http.HTTPStatus.OK, http.HTTPStatus.CREATED, http.HTTPStatus.ACCEPTED):
            logger.error(f"API Request: failed {response.status_code} {response.reason}")
            raise ConnectionError(f"{response.status_code} {response.reason}")
        else:
            return response

    def intent(self, imow_action: IMowActions, mower_name: str = "", mower_id: str = "", mower_action_id: str = "",
               startpoint: any = "0",
               duration: int = 30):
        """
        Intent to do a action. This seems to create a job object upstream. The action object contains an action Enum,
        the action Value is <MowerExternalId> or <MowerExternalId,DurationInMunitesDividedBy10,StartPoint> if
        startMowing is chosen

        :param imow_action: Anything from imow.common.actions
        :param mower_name: sth to identify which mower is used
        :param mower_id: sth to identify which mower is used
        :param mower_action_id:
            necessary identifier for the mowers for actions.
            This is looked up, if only mower_name or mower_id is provided

        :param startpoint: point from which the mowing shall start, default to 0
        :param duration: minutes of intended mowing defaults, to 30 minutes
        :return:
        """
        if not mower_action_id and not mower_id and not mower_name:
            raise AttributeError("Need some mower to work on. Please specify mower_[name|id|action_id]")
        if not mower_action_id and mower_name:
            mower_action_id = self.get_mower_action_id_from_name(mower_name)
        if not mower_action_id and mower_id:
            mower_action_id = self.get_mower_action_id_from_id(mower_id)

        if len(mower_action_id) < 16:
            raise AttributeError(
                f"Invalid mower_action_id, need exactly 16 chars, got {len(mower_action_id)} in {mower_action_id}")

        url = f"{IMOW_API_URI}/mower-actions/"

        action_value = f"{mower_action_id},{str(int(duration / 10))},{str(startpoint)}" \
            if imow_action == IMowActions.START_MOWING else mower_action_id

        action_object = {
            "actionName": imow_action.value,
            "actionValue": action_value
            # "0000000123456789,15,0" <MowerExternalId,DurationInMunitesDividedBy10,StartPoint>
        }
        logger.debug(f"Intend: {action_object}")

        payload = json.dumps(action_object)

        response = self.api_request(url, "POST", payload=payload)
        if not response.status_code == http.HTTPStatus.CREATED:
            raise ConnectionError(f"{response.status_code} {response.reason}")
        else:
            logger.debug(f'Sent mower {mower_action_id} to {imow_action}')
            return response

    def get_status_by_name(self, mower_name: str) -> dict:
        logger.debug(f"get_status_by_name: {mower_name}")
        for mower in self.receive_mowers():
            if mower.name == mower_name:
                return mower.status
        raise LookupError(f"Mower with name {mower_name} not found in upstream")

    def get_status_by_id(self, mower_id=(str, int)) -> dict:
        if not type(mower_id) == str:
            mower_id = str(mower_id)
        logger.debug(f"get_status_by_id: {mower_id}")
        try:
            return self.receive_mower_by_id(mower_id).status
        except ConnectionError:
            raise LookupError(f"Mower with id {mower_id} not found in upstream")

    def get_status_by_action_id(self, mower_action_id: str) -> dict:
        logger.debug(f"get_status_by_action_id: {mower_action_id}")
        for mower in self.receive_mowers():
            if mower.externalId == mower_action_id:
                return mower.status
        raise LookupError(f"Mower with externalId {mower_action_id} not found in upstream")

    def get_mower_action_id_from_name(self, mower_name: str) -> str:
        logger.debug(f"get_mower_action_id_from_name: {mower_name}")
        for mower in self.receive_mowers():
            if mower.name == mower_name:
                return mower.externalId
        raise LookupError(f"Mower with name {mower_name} not found in upstream")

    def get_mower_action_id_from_id(self, mower_id: str) -> str:
        logger.debug(f"get_mower_action_id_from_id: {mower_id}")
        try:
            return self.receive_mower_by_id(mower_id).externalId
        except ConnectionError:
            raise LookupError(f"Mower with id {mower_id} not found in upstream")

    def get_mower_id_from_name(self, mower_name: str) -> str:
        logger.debug(f"get_mower_id_from_name: {mower_name}")
        for mower in self.receive_mowers():
            if mower.name == mower_name:
                return mower.id
        raise LookupError(f"Mower with name {mower_name} not found in upstream")

    def receive_mowers(self) -> list[MowerState]:
        logger.debug(f"receive_mowers:")
        mowers = []
        for mower in json.loads(self.api_request(f"{IMOW_API_URI}/mowers/", "GET").text):
            mowers.append(MowerState(mower, self))
        logger.debug(mowers)
        return mowers

    def receive_mower_by_name(self, mower_name: str) -> MowerState:
        logger.debug(f"get_mower_from_name: {mower_name}")
        for mower in self.receive_mowers():
            if mower.name == mower_name:
                logger.debug(mower)
                return mower
        raise LookupError(f"Mower with name {mower_name} not found in upstream")

    def receive_mower_by_id(self, mower_id: str) -> MowerState:
        logger.debug(f"receive_mower: {mower_id}")
        mower = MowerState(
            json.loads(self.api_request(f"{IMOW_API_URI}/mowers/{mower_id}/", "GET").text), self)
        logger.debug(mower)
        return mower

    def receive_mower_current_task(self, mower_id: str) -> [MowerTask, int]:
        logger.debug(f"receive_mower_current_state: {mower_id}")
        state = MowerState(json.loads(
            self.api_request(f"{IMOW_API_URI}/mowers/{mower_id}/", "GET").text),
            self)
        logger.debug(state)
        try:
            return MowerTask(state.status.get("mainState"))
        except ValueError as ve:
            logger.warning(f'{state.status.get("mainState")} is not yet a known MowerTask-id to this Class.')
            return state.status.get("mainState")

    def receive_mower_statistics(self, mower_id: str) -> dict:
        logger.debug(f"receive_mower_statistics: {mower_id}")
        stats = json.loads(self.api_request(f"{IMOW_API_URI}/mowers/{mower_id}/statistic/", "GET").text)
        logger.debug(stats)
        return stats

    def receive_mower_week_mow_time_in_hours(self, mower_id: str) -> dict:
        logger.debug(f"receive_mower_week_mow_time_in_hours: {mower_id}")
        mow_times = json.loads(
            self.api_request(f"{IMOW_API_URI}/mowers/{mower_id}/statistics/week-mow-time-in-hours/",
                             "GET").text)
        logger.debug(mow_times)
        return mow_times

    def receive_mower_start_points(self, mower_id: str) -> dict:
        logger.debug(f"receive_mower_start_points: {mower_id}")
        start_points = json.loads(
            self.api_request(f"{IMOW_API_URI}/mowers/{mower_id}/start-points/", "GET").text)
        logger.debug(start_points)
        return start_points



import json
import logging

from imow.common.actions import IMowActions
from imow.common.consts import IMOW_API_URI
from imow.common.mowertask import MowerTask

logger = logging.getLogger('imow')


class MowerState:

    def __init__(self, upstream: dict, api):  # Type: api: IMowApi
        self.api = api
        self.update(upstream)

    def update(self, upstream: dict = None):
        if not upstream:
            upstream = json.loads(self.api.api_request(f"{IMOW_API_URI}/mowers/{self.id}/", "GET").text)
        self.__dict__.update(map(lambda kv: (kv[0].replace(' ', '_'), kv[1]), upstream.items()))

    def get_current_task(self) -> (MowerTask, int):
        return self.api.receive_mower_current_task(mower_id=self.id)

    def intent(self, imow_action: IMowActions, startpoint: any = "0", duration: int = 30):
        self.api.intent(imow_action=imow_action, startpoint=startpoint, duration=duration,
                        mower_action_id=self.externalId)

    def get_status(self) -> dict:
        self.update()
        return self.status

    def get_statistics(self) -> dict:
        return self.api.receive_mower_statistics(self.id)

    def get_startpoints(self) -> dict:
        return self.api.receive_mower_start_points(self.id)

    def get_mower_week_mow_time_in_hours(self) -> dict:
        return self.api.receive_mower_week_mow_time_in_hours(self.id)

    accountId: str = {str}
    asmEnabled: bool = {bool}
    automaticModeEnabled: bool = {bool}
    boundryOffset: bool = {int}  # 60
    childLock: bool = {bool}  # False
    circumference: bool = {int}  # 41
    cModuleId: str = {str}  # '0234d0fffab1d345'
    codePage: bool = {int}  # 0
    coordinateLatitude: float = {float}  # 54.123456
    coordinateLongitude: float = {float}  # 10.654321
    corridorMode: bool = {int}  # 0
    demoModeEnabled: bool = {bool}  # False
    deviceType: bool = {int}  # 24
    deviceTypeDescription: str = {str}  # 'RMI 422 PC'
    edgeMowingMode: bool = {int}  # 2
    endOfContract: str = {str}  # '2000-01-01T01:59:43+00:00'
    energyMode: bool = {int}  # 3
    externalId: str = {str}  # '0000000123456789'
    firmwareVersion: str = {str}  # '01v013'
    gdprAccepted: bool = {bool}  # True
    gpsProtectionEnabled: bool = {bool}  # True
    id: str = {str}  # '31466'
    imsi: str = {str}  # '16198732186461'
    lastWeatherCheck: str = {str}  # '2021-05-15T01:42:25+00:00'
    ledStatus: bool = {int}  # 11
    localTimezoneOffset: bool = {int}  # 7182
    mappingIntelligentHomeDrive: bool = {int}  # 0
    mowerImageThumbnailUrl = {
        str}  # 'https://app-cdn-appdata001-r-euwe-1b3d32.azureedge.net/device-images/mower-images/31466-2309868077
    # -thumb.png'
    mowerImageUrl = {
        str}  # 'https://app-cdn-appdata001-r-euwe-1b3d32.azureedge.net/device-images/mower-images/31466-2309868077
    # -photo.png'
    name: str = {str}  # 'Mährlin'
    protectionLevel: bool = {int}  # 1
    rainSensorMode: bool = {int}  # 1
    smartLogic: dict = {dict: 13}
    # {'dynamicMowingplan': False, 'mower': None, 'mowingArea': 100, 'mowingAreaInFeet': 1000, 'mowingAreaInMeter': 100,
    # 'mowingGrowthAdjustment': 0, 'mowingTime': 60, 'mowingTimeManual': False, 'performedActivityTime': 3,
    # 'smartNotifications': False, 'suggestedActivityTime': 135, 'totalActivityActiveTime': 0,
    # 'weatherForecastEnabled': True}
    softwarePacket: str = {str}  # '12.03'
    status: dict = {dict: 15}
    # {'bladeService': False, 'chargeLevel': 66, 'extraStatus': 0, 'extraStatus1': 0, 'extraStatus2': 0,
    # 'extraStatus3': 0, 'extraStatus4': 0, 'extraStatus5': 0, 'lastGeoPositionDate': '2021-05-15T07:12:10+00:00',
    # 'lastNoErrorMainState': 7, 'lastSeenDate': '2021-05-13T23:58:31+00:00', 'mainState': 7, 'mower': None,
    # 'online': True, 'rainStatus': False}
    team = {None}  # None
    teamable: bool = {bool}  # False
    timeZone: str = {str}  # 'Europe/Berlin'
    unitFormat: bool = {int}  # 0
    version: str = {str}  # '3.2.038'



import re
from setuptools import find_packages, setup

with open("imow/common/package_descriptions.py", encoding='utf-8') as file_handler:
    lines = file_handler.read()
    version = re.search(r'__version__ = "(.*?)"', lines).group(1)
    package_name = re.search(r'package_name = "(.*?)"', lines).group(1)
    python_major = int(re.search(r'python_major = "(.*?)"', lines).group(1))
    python_minor = int(re.search(r'python_minor = "(.*?)"', lines).group(1))
PACKAGES = [f"imow.{p}" for p in find_packages(where="imow")]
with open("requirements/release.txt", mode='r', encoding='utf-8') as requirements:
    packages = requirements.read().splitlines()

setup(
    name="imow-webapi",
    version=version,
    author="ChrisHaPunkt",
    description="A library to authenticate and interact with STIHL iMow mowers using their WebAPI",
    long_description=open("README.md").read() + "\n\n" + open("CHANGELOG.md").read(),
    long_description_content_type="text/markdown",
    license="GPL",
    keywords="stihl imow mower api",
    url="https://github.com/ChrisHaPunkt/stihl-imow-webapi",
    packages=PACKAGES,
    namespace_packages=["imow"],
    zip_safe=False,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    test_suite="tests",
    install_requires=packages,
    setup_requires=["pytest-runner"],
    entry_points={
        'console_scripts': ['%s=%s.__init__:main' % (package_name, package_name)]
    },
)

wheel_name = package_name.replace('-', '_') if '-' in package_name else package_name
print("Setup is complete. Run 'python -m pip install dist/%s-%s-py%d-none-any.whl' to install this wheel." % (wheel_name, version, python_major))
